@@ -4,15 +4,21 @@
 #'
 #' @importFrom purrr invoke
 #' @importFrom mc2d rpert
+#' @importFrom stringi stri_split_fixed
 #' @param func Function to use to simulate TEF, defaults to \code{\link[mc2d]{rpert}}.
 #' @param params Optional parameters to pass to `func`.
 #' @return List containing type ("tef"), samples (as a vector), and details (as a list).
 #' @family OpenFAIR helpers
 #' @export
 sample_tef <- function(func = NULL, params = NULL) {
-  if (is.null(func)) func <- get("rpert", asNamespace("mc2d"))
+  if (is.null(func)) func <- get("rpert", asNamespace("mc2d")) else {
+    func_split <- stringi::stri_split_fixed(func, "::", simplify = TRUE)
+    requireNamespace(func_split[1], quietly = TRUE)
+    func <- get(func_split[2], asNamespace(func_split[1]))
+  }
+
   list(type = "tef",
-       samples = as.integer(round(invoke(func, params))),
+       samples =  as.integer(round(purrr::invoke(func, params))),
        details = list())
 }
 
@@ -20,15 +26,20 @@ sample_tef <- function(func = NULL, params = NULL) {
 #'
 #' @importFrom purrr invoke
 #' @importFrom mc2d rpert
+#' @importFrom stringi stri_split_fixed
 #' @param func Function to use to simulate TC, defaults to \code{\link[mc2d]{rpert}}.
 #' @param params Optional parameters to pass to `func`.
 #' @return List containing type ("tc"), samples (as a vector), and details (as a list).
 #' @family OpenFAIR helpers
 #' @export
 sample_tc <- function(func = NULL, params = NULL) {
-  if (is.null(func)) func <- get("rpert", asNamespace("mc2d"))
+  if (is.null(func)) func <- get("rpert", asNamespace("mc2d")) else {
+    func_split <- stringi::stri_split_fixed(func, "::", simplify = TRUE)
+    requireNamespace(func_split[1], quietly = TRUE)
+    func <- get(func_split[2], asNamespace(func_split[1]))
+  }
   list(type = "tc",
-       samples = invoke(func, params),
+       samples = if (params["n"] != 0) purrr::invoke(func, params) else NA,
        details = list())
 }
 
@@ -37,13 +48,19 @@ sample_tc <- function(func = NULL, params = NULL) {
 #'
 #' @importFrom purrr invoke
 #' @importFrom mc2d rpert
+#' @importFrom stringi stri_split_fixed
 #' @param func Function to use to simulate DIFF, defaults to \code{\link[mc2d]{rpert}}.
 #' @param params Optional parameters to pass to `func`.
 #' @return List containing type ("diff"), samples (as a vector), and details (as a list).
 #' @family OpenFAIR helpers
 #' @export
 sample_diff <- function(func = NULL, params = NULL) {
-  if (is.null(func)) func <- get("rpert", asNamespace("mc2d"))
+  if (is.null(func)) func <- get("rpert", asNamespace("mc2d")) else {
+    func_split <- stringi::stri_split_fixed(func, "::", simplify = TRUE)
+    requireNamespace(func_split[1], quietly = TRUE)
+    func <- get(func_split[2], asNamespace(func_split[1]))
+  }
+
   list(type = "diff",
        samples = invoke(func, params),
        details = list())
@@ -52,14 +69,20 @@ sample_diff <- function(func = NULL, params = NULL) {
 #' Calculate the vulnerability
 #'
 #' @importFrom purrr invoke is_list
+#' @importFrom stringi stri_split_fixed
 #' @param func Function to use to simulate VULN, defaults to \code{\link[stats]{rbinom}}.
 #' @param params Optional parameters to pass to `func`.
 #' @return List containing type ("vuln"), samples (as a vector), and details (as a list).
 #' @family OpenFAIR helpers
 #' @export
 sample_vuln <- function(func = NULL, params = NULL) {
-  if (is.null(func)) func <- get("rbinom", asNamespace("stats"))
-  dat <- invoke(func, params)
+  if (is.null(func)) func <- get("rbinom", asNamespace("stats")) else {
+    func_split <- stringi::stri_split_fixed(func, "::", simplify = TRUE)
+    requireNamespace(func_split[1], quietly = TRUE)
+    func <- get(func_split[2], asNamespace(func_split[1]))
+  }
+
+  dat <- purrr::invoke(func, params)
   list(type = "vuln",
        samples = if (purrr::is_list(dat)) dat$samples else dat,
        details = if (purrr::is_list(dat)) dat$details else list()
@@ -69,6 +92,7 @@ sample_vuln <- function(func = NULL, params = NULL) {
 #' Given a number of loss events and a loss distribution, calculate losses
 #'
 #' @importFrom purrr invoke
+#' @importFrom stringi stri_split_fixed
 #' @importFrom mc2d rpert
 #' @param func Function to use to simulate TEF, defaults to \code{\link[mc2d]{rpert}}.
 #' @param params Optional parameters to pass to `func`.
@@ -77,8 +101,12 @@ sample_vuln <- function(func = NULL, params = NULL) {
 #' @export
 sample_lm <- function(func = NULL, params = NULL) {
 
-  if (is.null(func)) func <- get("rpert", asNamespace("mc2d"))
-  samples <- invoke(func, params)
+  if (is.null(func)) func <- get("rpert", asNamespace("mc2d")) else {
+    func_split <- stringi::stri_split_fixed(func, "::", simplify = TRUE)
+    requireNamespace(func_split[1], quietly = TRUE)
+    func <- get(func_split[2], asNamespace(func_split[1]))
+  }
+  samples <- if (params["n"] != 0 & !is.na(params["n"])) purrr::invoke(func, params) else 0
 
   # We have to calculate ALE/SLE differently (ALE: 0, SLE: NA) if there are no losses
   details <- if (length(samples) == 0 | sum(samples) == 0) {
@@ -106,8 +134,13 @@ sample_lm <- function(func = NULL, params = NULL) {
 #' @family OpenFAIR helpers
 #' @export
 sample_lef <- function(func = NULL, params = NULL) {
-  if (is.null(func)) func <- get("rnorm", asNamespace("stats"))
-  dat <- invoke(func, params)
+  if (is.null(func)) func <- get("rpert", asNamespace("mc2d")) else {
+    func_split <- stringi::stri_split_fixed(func, "::", simplify = TRUE)
+    requireNamespace(func_split[1], quietly = TRUE)
+    func <- get(func_split[2], asNamespace(func_split[1]))
+  }
+
+  dat <- purrr::invoke(func, params)
   list(type = "lef",
        samples = if (purrr::is_list(dat)) dat$samples else dat,
        details = if (purrr::is_list(dat)) dat$details else list()
@@ -119,30 +152,41 @@ sample_lef <- function(func = NULL, params = NULL) {
 #' Calculate difficulty strength across multiple controls by taking the mean
 #'
 #' Given a set of estimation parameters, calculate control strength as the
-#' arithmetic mean of sampled control effectiveness.
+#'   arithmetic mean of sampled control effectiveness.
 #'
-#' @importFrom dplyr %>%
+#' @importFrom dplyr %>% bind_rows
 #' @importFrom purrr pmap map transpose simplify_all map_dbl
+#' @importFrom tidyr nest
+#' @importFrom stringi stri_split_fixed
+#' @importFrom rlang .data
 #' @param n Number of threat events to sample controls across.
-#' @param diff_estimates Parameters to pass to \code{\link{sample_diff}}.
+#' @param diff_parameters Parameters to pass to \code{\link{sample_diff}}.
 #' @return Vector of control effectiveness.
 #' @family OpenFAIR helpers
 #' @export
-get_mean_control_strength <- function(n, diff_estimates)  {
-  # ensure control estimates are in the order we expect
-  control_params <- with(diff_estimates, list(l = l, ml = ml, h = h,
-                                              conf = conf))
-  cs <- purrr::pmap(control_params, ~ sample_diff(
-    params = list(n, ..1, ..2, ..3, ..4))) %>%
-    purrr::map("samples")
-  cs <- cs %>% purrr::transpose(.) %>%  purrr::simplify_all(.)
-  # take the mean of all controls
+get_mean_control_strength <- function(n, diff_parameters)  {
+  # get the list of control parameters, including the function to call
+  # diff_parameters is a list of lists, so strip one layer to get a simple list
+  control_list <- diff_parameters %>% purrr::flatten()
+
+  # iterate over the control parameters list
+  # getting a number of samples for each control
+  cs <- purrr::map(control_list, function(x) {
+    # create a nested tibble (func, params)
+    diff_tbl <- x %>% dplyr::bind_rows() %>%
+      tidyr::nest(-.data$func, .key = "params")
+    # extract the parameters list from the tibble
+    params <- c(n = n, diff_tbl$params %>% unlist())
+    #generate the samples
+    sample_diff(func = diff_tbl$func, params = params)
+    }) %>% purrr::map("samples")
+
+  # pivot the results so each list has 1 sample for each control
+  cs <- purrr::transpose(cs) %>% purrr::simplify_all()
+
+  # take the mean of each list, giving the mean control strength for that event
   outcomes <- cs %>% purrr::map_dbl(mean)
-  # placeholder for control importance work
-  # cs_df <- purrr::map_dfc(cs, tibble::as_tibble) %>% t %>% tibble::as_tibble
-  # control_importance <- caret::filterVarImp(cs_df, outcomes, nonpara = TRUE) %>%
-  #   tibble::rownames_to_column(var = "control") %>%
-  #   arrange(desc(Overall)) %>% dplyr::pull("control")
+
   outcomes
 }
 
@@ -151,8 +195,8 @@ get_mean_control_strength <- function(n, diff_estimates)  {
 #' Calculate number of loss events which occur in a simulated period
 #'
 #' Composition function for use in \code{\link{sample_lef}}. Given a count of
-#' the number of threat events (TEF) and the level of vulnerability (as a
-#' percentage), calculate how many of those become loss events (LEF).
+#'   the number of threat events (TEF) and the level of vulnerability (as a
+#'   percentage), calculate how many of those become loss events (LEF).
 #'
 #' @param tef Threat event frequency (n).
 #' @param vuln Vulnerability (percentage).
@@ -170,8 +214,8 @@ compare_tef_vuln <- function(tef, vuln) {
 #' Determine which threat events result in loss opportunities
 #'
 #' Composition function for use in \code{\link{sample_vuln}}, does a simple
-#' compare of all threat events where the threat capability (TC) is greater
-#' than the difficulty (DIFF).
+#'   compare of all threat events where the threat capability (TC) is greater
+#'   than the difficulty (DIFF).
 #'
 #' @param tc Threat capability (as a percentage).
 #' @param diff Difficulty (as a percentage).
@@ -186,13 +230,19 @@ select_loss_opportunities <- function(tc, diff) {
   samples <-  tc > diff
 
   # mean amount threat strength exceeds control strength, if that ever occurs
-  tc_exceedance <- if (sum(samples) > 0) {
-    mean(tc[samples] - diff[samples], na.rm = TRUE)
-    } else {0}
+  tc_exceedance <- if (
+    all(is.na(samples))) {NA} else # if there are no threat events, then NA
+      if (sum(samples, na.rm = TRUE) > 0) {
+        mean(tc[samples] - diff[samples], na.rm = TRUE)
+      } else {0}
   # mean amount control strength exceeds threat strength, if that ever occurs
-  diff_exceedance <- if (sum(samples) != length(tc)) {
-    mean(diff[!samples] - tc[!samples], na.rm = TRUE)
-    } else {0}
+  diff_exceedance <- if (
+    all(is.na(samples))) {NA} else # if there are no threat events, then NA
+      if (
+        #sum(samples, na.rm = TRUE) > 0 &
+          sum(samples, na.rm = TRUE) != length(tc)) {
+        mean(diff[!samples] - tc[!samples], na.rm = TRUE)
+        } else {0}
   list(samples = samples, details = list(mean_tc_exceedance = tc_exceedance,
                                          mean_diff_exceedance = diff_exceedance))
 }
@@ -202,15 +252,16 @@ select_loss_opportunities <- function(tc, diff) {
 #' Run an OpenFAIR simulation at the TEF/TC/DIFF/LM levels
 #'
 #' Run an OpenFAIR model with parameters provided for TEF, TC, DIFF, and
-#' LM sampling. If there are multiple controls provided for a scenarios, the
-#' arithmetic mean (average) is taken across samples for all controls to get
-#' the effective control strength for a given simulation.
+#'   LM sampling. If there are multiple controls provided for a scenarios, the
+#'   arithmetic mean (average) is taken across samples for all controls to get
+#'   the effective control strength for a given simulation.
 #'
-#' @importFrom purrr pmap map pluck simplify_all transpose map_dbl map_int
-#' @importFrom tibble tibble
+#' @importFrom purrr pmap map pluck simplify_all transpose map_dbl map_int flatten
+#' @importFrom tidyr nest
+#' @importFrom tibble tibble as_tibble
+#' @importFrom rlang .data
 #' @importFrom dplyr %>%
-#' @param scenario List of tef_, tc_, and LM_ l/ml/h/conf parameters.
-#' @param diff_estimates Parameters for estimating the scenario difficulty.
+#' @param scenario List of TEF, TC, and LM parameters.
 #' @param n Number of simulations to run.
 #' @param title Optional name of scenario.
 #' @param verbose Whether to print progress indicators.
@@ -221,10 +272,9 @@ select_loss_opportunities <- function(tc, diff) {
 #' @examples
 #' data(quantitative_scenarios)
 #' scenario <- quantitative_scenarios[1, ]
-#' controls <- scenario[[1, "diff_params"]]
-#' openfair_tef_tc_diff_lm(scenario, controls, 10)
-openfair_tef_tc_diff_lm <- function(scenario, diff_estimates, n = 10^4,
-                                    title = "Untitled", verbose = FALSE) {
+#' openfair_tef_tc_diff_lm(scenario, 10)
+openfair_tef_tc_diff_lm <- function(scenario, n = 10^4, title = "Untitled",
+                                    verbose = FALSE) {
 
     # make samples repeatable (and l33t)
     set.seed(31337)
@@ -236,22 +286,21 @@ openfair_tef_tc_diff_lm <- function(scenario, diff_estimates, n = 10^4,
     }
 
     # TEF - how many contacts do we have in each simulated period
-    TEFestimate <- with(scenario, tibble::tibble(l = tef_l, ml = tef_ml,
-                                                 h = tef_h, conf = tef_conf))
-    TEFsamples <- sample_tef(params = list(n, TEFestimate$l, TEFestimate$ml,
-                                           TEFestimate$h,
-                                           shape = TEFestimate$conf))
+    TEFestimate <- scenario$tef_params %>% purrr::flatten() %>%
+      tibble::as_tibble() %>% tidyr::nest(-.data$func, .key = "params")
+    params <- c(n = n, TEFestimate$params %>% unlist())
+    TEFsamples <- sample_tef(func = TEFestimate$func, params = params)
     TEFsamples <- TEFsamples$samples
 
     # TC - what is the strength of each threat event
     #    - get the threat capability parameters for this scenario
-    TCestimate <- with(scenario, tibble::tibble(l = tc_l, ml = tc_ml,
-                                                h = tc_h, conf = tc_conf))
+    TCestimate <- scenario$tc_params %>% purrr::flatten() %>%
+      tibble::as_tibble() %>% tidyr::nest(-.data$func, .key = "params")
     #    - sample threat capability for each TEF event in each sample period
-    TCsamples <- purrr::map(1:n, ~ sample_tc(params = list(TEFsamples[.x], TCestimate$l,
-                                                   TCestimate$ml,
-                                                   TCestimate$h,
-                                                   shape = TCestimate$conf))$samples)
+    TCsamples <- purrr::map(1:n, function(x) {
+      params <- c(n = TEFsamples[x], TCestimate$params %>% unlist())
+      sample_tc(func = TCestimate$func, params = params)
+      })
     # TCSamples is now a list of of the TC for each threat event
 
     # DIFF - calculate the mean strength of controls for each threat event
@@ -260,7 +309,7 @@ openfair_tef_tc_diff_lm <- function(scenario, diff_estimates, n = 10^4,
     # get the difficulty for each threat event across all the simulated periods
     DIFFsamples <- purrr::map(1:n, function(x) {
       if (is.numeric(TEFsamples[[x]]) && TEFsamples[[x]] > 0) {
-        get_mean_control_strength(TEFsamples[[x]], diff_estimates)
+        get_mean_control_strength(TEFsamples[[x]], scenario$diff_params)
         } else {NA}
       })
     # DIFFsamples is now a list of vectors of the control strength for
@@ -268,8 +317,8 @@ openfair_tef_tc_diff_lm <- function(scenario, diff_estimates, n = 10^4,
 
     # LEF - determine how many threat events become losses (TC > DIFF)
     LEFsamples <- purrr::map(1:n, function(x) {
-      sample_lef(func = select_loss_opportunities,
-                 params = list(tc = TCsamples[[x]],
+      sample_lef(func = "evaluator::select_loss_opportunities",
+                 params = list(tc = TCsamples[[x]]$samples,
                                diff = DIFFsamples[[x]]))
     })
 
@@ -278,11 +327,11 @@ openfair_tef_tc_diff_lm <- function(scenario, diff_estimates, n = 10^4,
     LEFsamples <- purrr::map(LEFsamples, c("samples")) %>% purrr::map_int(sum)
 
     # LM - determine the size of losses for each simulation
-    LMestimate <- with(scenario, tibble::tibble(l = lm_l, ml = lm_ml,
-                                                h = lm_h, conf = lm_conf))
+    LMestimate <- scenario$lm_params %>% purrr::flatten() %>%
+      tibble::as_tibble() %>% tidyr::nest(-.data$func, .key = "params")
     loss_samples <- purrr::map(LEFsamples, function(x) {
-      dat <- sample_lm(params = list(x, LMestimate$l, LMestimate$ml,
-                                    LMestimate$h, LMestimate$conf))
+      params <- c(n = x, LMestimate$params %>% unlist())
+      dat <- sample_lm(func = LMestimate$func, params = params)
       dat$samples <- sum(dat$samples)
       dat
       })
@@ -290,8 +339,8 @@ openfair_tef_tc_diff_lm <- function(scenario, diff_estimates, n = 10^4,
     # summary stats for ALE
     if (verbose) {
         print(summary(purrr::map_dbl(loss_samples, "samples")))
-        value_at_risk <- quantile(purrr::map_dbl(loss_samples, "samples"),
-                                  probs = (0.95), na.rm = TRUE)
+        value_at_risk <- stats::quantile(purrr::map_dbl(loss_samples, "samples"),
+                                         probs = (0.95), na.rm = TRUE)
         message(paste0("Losses at 95th percentile are $",
                        format(value_at_risk, nsmall = 2, digits = 2,
                               big.mark = ",")))
