@@ -10,13 +10,6 @@ test_that("Dollar Millions formats as expected", {
   expect_equal(dollar_millions(1.523 * 10^6), "$1.52M")
 })
 
-test_that("Domain TC vs DIFF calculations", {
-  data("simulation_results")
-  data("domains")
-  dat <- calculate_weak_domains(simulation_results, domains)
-  expect_equal(nrow(dat), nrow(domains))
-})
-
 test_that("Calculate max losses", {
   data("simulation_results")
   dat <- calculate_max_losses(simulation_results, c(1, 10))
@@ -25,10 +18,9 @@ test_that("Calculate max losses", {
   expect_equal(nrow(dat), 2000)
 })
 
-test_that("Calculate max losses", {
+test_that("calculate_max_losses handles NULL outliers", {
   data("simulation_results")
-  data("domains")
-  dat <- calculate_domain_impact(domain_summary, domains)
+  dat <- calculate_max_losses(simulation_results)
   expect_s3_class(dat, "data.frame")
-  expect_equal(nrow(dat), nrow(domains))
+  expect_equal(nrow(dat), 1000)
 })
